# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import collections

from botocore.exceptions import (
    SSOError,
    SSOTokenLoadError,
    UnauthorizedSSOTokenError,
)

from .vendored_botocore.exceptions import PendingAuthorizationExpiredError

class InvalidSSOConfigError(Exception):
    pass

class AuthDispatchError(Exception):
    pass

class AuthenticationNeededError(Exception):
    pass



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os
import sys
import textwrap
import webbrowser

from .exceptions import AuthenticationNeededError, AuthDispatchError

DEFAULT_BROWSER_MESSAGE = textwrap.dedent("""\
AWS SSO login required.
Attempting to open the SSO authorization page in your default browser.
If the browser does not open or you wish to use a different device to
authorize this request, open the following URL:

{url}

Then enter the code:

{code}
""")

DEFAULT_NO_BROWSER_MESSAGE = textwrap.dedent("""\
AWS SSO login required.
Open the following URL in a browser:

{url}

Then enter the code:

{code}
""")

class OpenBrowserHandler(object):
    def __init__(self, outfile=None, open_browser=None, disable_browser=None, message=None):
        if outfile is None:
            outfile = sys.stderr
        self._outfile = outfile

        if open_browser is None:
            open_browser = webbrowser.open_new_tab
        self._open_browser = open_browser

        if disable_browser is None:
            disable_browser = os.environ.get('AWS_SSO_DISABLE_BROWSER', '').lower() in ['1', 'true']
        self._disable_browser = disable_browser

        if message is None:
            if self._disable_browser:
                message = DEFAULT_NO_BROWSER_MESSAGE
            else:
                message = DEFAULT_BROWSER_MESSAGE
        self._message = message

    def __call__(self, userCode, verificationUri,
                 verificationUriComplete, **kwargs):

        if self._message and self._outfile:
            message = self._message.format(
                url=verificationUri,
                code=userCode,
                verificationUri=verificationUri,
                userCode=userCode
            )

            print(message, file=self._outfile)

        if self._open_browser and not self._disable_browser:
            try:
                return self._open_browser(verificationUriComplete)
            except AuthenticationNeededError:
                raise
            except Exception as e:
                raise AuthDispatchError('Failed to open browser') from e
                # LOG.debug('Failed to open browser:', exc_info=True)

def non_interactive_auth_raiser(*args, **kwargs):
    raise AuthenticationNeededError



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This code is based on the code for the AWS CLI v2"s `aws sso login` functionality
# https://github.com/aws/aws-cli/tree/v2/awscli/customizations/sso

import os
import sys
import webbrowser
import logging
import datetime
import uuid
import numbers
import typing

import boto3
import botocore
from botocore.credentials import JSONFileCache
from botocore.credentials import SSOCredentialFetcher

from .format import format_account_id

from .vendored_botocore.utils import SSOTokenFetcher

from .exceptions import InvalidSSOConfigError, AuthDispatchError, AuthenticationNeededError
from .browser import OpenBrowserHandler, non_interactive_auth_raiser

SSO_TOKEN_DIR = os.path.expanduser(
    os.path.join("~", ".aws", "sso", "cache")
)

CREDENTIALS_CACHE_DIR = os.path.expanduser(
    os.path.join("~", ".aws", "cli", "cache")
)

LOGGER = logging.getLogger(__name__)

__all__ = ["get_boto3_session", "login", "list_available_accounts", "list_available_roles"]

def get_token_fetcher(session, sso_region, *, interactive=False, sso_cache=None,
                     on_pending_authorization=None, message=None, outfile=None,
                     disable_browser=None):
    if hasattr(session, "_session"): #boto3 Session
        session = session._session

    if sso_cache is None:
        sso_cache = JSONFileCache(SSO_TOKEN_DIR)

    if on_pending_authorization is None:
        if interactive:
            on_pending_authorization = OpenBrowserHandler(
                outfile=outfile,
                message=message,
                disable_browser=disable_browser,
            )
        else:
            on_pending_authorization = non_interactive_auth_raiser

    token_fetcher = SSOTokenFetcher(
        sso_region=sso_region,
        client_creator=session.create_client,
        cache=sso_cache,
        on_pending_authorization=on_pending_authorization,
    )
    return token_fetcher

def _get_token_loader_from_token_fetcher(token_fetcher, force_refresh=False):
    def token_loader(start_url):
        token_response = token_fetcher.fetch_token(
            start_url=start_url,
            force_refresh=force_refresh
        )
        LOGGER.debug("TOKEN: {}".format(token_response))
        return token_response["accessToken"]

    return token_loader


def get_credentials(session, start_url, sso_region, account_id, role_name, *,
        token_fetcher=None,
        force_refresh=False,
        sso_cache=None,
        credential_cache=None):
    """Return credentials for the given role.

    The return value is a dict containing the assumed role credentials.

    You probably want to use get_boto3_session() instead, which returns
    a boto3 session that caches its credentials and automatically refreshes them.
    """
    if hasattr(session, "_session"): #boto3 Session
        session = session._session
    if not token_fetcher:
        token_fetcher = get_token_fetcher(session, sso_region, sso_cache=sso_cache)

    token_loader = _get_token_loader_from_token_fetcher(token_fetcher=token_fetcher, force_refresh=force_refresh)

    if credential_cache is None:
        credential_cache = JSONFileCache(CREDENTIALS_CACHE_DIR)

    credential_fetcher = SSOCredentialFetcher(
        start_url=start_url,
        sso_region=sso_region,
        role_name=role_name,
        account_id=account_id,
        client_creator=session.create_client,
        cache=credential_cache,
        token_loader=token_loader,
    )

    return credential_fetcher.fetch_credentials()

def _get_botocore_session(
        start_url,
        sso_region,
        account_id,
        role_name,
        credential_cache=None,
        sso_cache=None,
        ):
    botocore_session = botocore.session.Session()

    profile_name = str(uuid.uuid4())
    load_config = lambda: {
        "profiles": {
            profile_name: {
                "sso_start_url": start_url,
                "sso_region": sso_region,
                "sso_account_id": account_id,
                "sso_role_name": role_name,
            }
        }
    }
    sso_provider = botocore.credentials.SSOProvider(
        load_config=load_config,
        client_creator=botocore_session.create_client,
        profile_name=profile_name,
        cache=credential_cache,
        token_cache=sso_cache,
    )

    botocore_session.register_component(
        "credential_provider",
        botocore.credentials.CredentialResolver([sso_provider])
    )

    return botocore_session

def get_boto3_session(
        start_url: str,
        sso_region: str,
        account_id: typing.Union[str, int],
        role_name: str,
        *,
        region: str,
        login: bool=False,
        sso_cache=None,
        credential_cache=None) -> boto3.Session:
    """Get a boto3 session with the input configuration.

    Args:
        start_url (str): The start URL for the AWS SSO instance.
        sso_region (str): The AWS region for the AWS SSO instance.
        account_id (str): The AWS account ID to use.
        role_name (str): The AWS SSO role (aka Permission Set) name to use.
        region (str): The AWS region for the boto3 session.
        login (bool): Interactively log in the user if their AWS SSO credentials have expired.
        sso_cache: A dict-like object for AWS SSO credential caching.
        credential_cache: A dict-like object to cache the role credentials in.

    Returns:
        A boto3 Session object configured for the account and role.
    """
    account_id = format_account_id(account_id)

    if login:
        _login(start_url, sso_region, sso_cache=sso_cache)

    botocore_session = _get_botocore_session(start_url, sso_region, account_id, role_name,
        credential_cache=credential_cache,
        sso_cache=sso_cache)

    session = boto3.Session(botocore_session=botocore_session, region_name=region)

    return session

def login(
        start_url: str,
        sso_region: str,
        *,
        force_refresh: bool=False,
        disable_browser: bool=None,
        message: str=None,
        outfile: typing.Union[typing.TextIO, typing.Literal[False]]=None,
        sso_cache=None) -> typing.Dict:
    """Interactively log in the user if their AWS SSO credentials have expired.

    If the user is not logged in or force_refresh is True, it will attempt to log in.
    If the user is logged in and force_refresh is False, no action is taken.

    If disable_browser is True, a message will be printed to stderr
    with a URL and code for the user to log in with.
    Otherwise, it will attempt to automatically open the user's browser
    to log in, as well as printing the URL and code to stderr as a fallback.

    A custom message can be printed by setting message to a template string
    using {url} and {code} as placeholders.
    The message can be suppressed by setting message to False.

    Args:
        start_url (str): The start URL for the AWS SSO instance.
        sso_region (str): The AWS region for the AWS SSO instance.
        force_refresh (bool): Always go through the authentication process.
        disable_browser (bool): Skip the browser popup and only print a message with the URL and code.
        message (str): A message template to print with the fallback URL and code.
        outfile (file): The file-like object to print the message to,
            or False to suppress the message.
        sso_cache: A dict-like object for AWS SSO credential caching.

    Returns:
        The token dict as returned by sso-oidc:CreateToken,
        which contains the actual authorization token, as well as the expiration.
        """
    session = botocore.session.Session()

    token_fetcher = get_token_fetcher(
        session=session,
        sso_region=sso_region,
        interactive=True,
        message=message,
        outfile=outfile,
        disable_browser=disable_browser,
        sso_cache=sso_cache)

    token = token_fetcher.fetch_token(
        start_url=start_url,
        force_refresh=force_refresh
    )

    return token

_login = login

def logout(
        start_url: str,
        sso_region: str,
        *,
        sso_cache=None):
    """Log out of the given AWS SSO instance.

    Note that this function currently does not remove the token from
    the SSO file cache, which can cause subsequent usage of other functions
    in this module to pick up an invalid token from there.
    https://github.com/boto/botocore/issues/2255

    Args:
        start_url (str): The start URL for the AWS SSO instance.
        sso_region (str): The AWS region for the AWS SSO instance.
        sso_cache: A dict-like object for AWS SSO credential caching.

    Returns:
        Never raises.
        Returns True if a token was found and successfully logged out.
        Returns False if no token was found.
        If any exception is raised during the logout process,
            it is caught and returned.
    """

    session = botocore.session.Session()

    token_fetcher = get_token_fetcher(
        session=session,
        sso_region=sso_region,
        sso_cache=sso_cache)

    try:
        token = token_fetcher.pop_token_from_cache(
            start_url=start_url
        )

        if not token:
            return True
        else:
            config = botocore.config.Config(
                region_name=sso_region,
                signature_version=botocore.UNSIGNED,
            )
            client = session.create_client("sso", config=config)

            client.logout(accessToken=token["accessToken"])
            return False
    except Exception as e:
        LOGGER.debug("Exception during logout", exc_info=True)
        return e

def list_available_accounts(
        start_url: str,
        sso_region: str,
        *,
        login: bool=False,
        sso_cache=None) -> typing.Iterator[typing.Tuple[str, str]]:
    """Iterate over the available accounts the user has access to through AWS SSO.

    Args:
        start_url (str): The start URL for the AWS SSO instance.
        sso_region (str): The AWS region for the AWS SSO instance.
        login (bool): Interactively log in the user if their AWS SSO credentials have expired.
        sso_cache: A dict-like object for AWS SSO credential caching.

    Returns:
        An iterator that yields account id and account name.
    """
    session = botocore.session.Session()

    token_fetcher = get_token_fetcher(session, sso_region, interactive=login, sso_cache=sso_cache)

    token = token_fetcher.fetch_token(start_url)

    config = botocore.config.Config(
        region_name=sso_region,
        signature_version=botocore.UNSIGNED,
    )
    client = session.create_client("sso", config=config)

    list_accounts_args = {"accessToken": token["accessToken"]}
    while True:
        response = client.list_accounts(**list_accounts_args)

        for account in response["accountList"]:
            yield account["accountId"], account["accountName"]

        next_token = response.get("nextToken")
        if not next_token:
            break
        else:
            list_accounts_args["nextToken"] = response["nextToken"]

def list_available_roles(
        start_url: str,
        sso_region: str,
        account_id: typing.Union[str, int, typing.Iterable[typing.Union[str, int]]]=None,
        *,
        login: bool=False,
        sso_cache=None) -> typing.Iterator[typing.Tuple[str, str, str]]:
    """Iterate over the available accounts and roles the user has access to through AWS SSO.

    Args:
        start_url (str): The start URL for the AWS SSO instance.
        sso_region (str): The AWS region for the AWS SSO instance.
        account_id: Optional account id or list of account ids to check.
            If not set, all accounts available to the user are listed.
        login (bool): Interactively log in the user if their AWS SSO credentials have expired.
        sso_cache: A dict-like object for AWS SSO credential caching.

    Returns:
        An iterator that yields account id, account name, and role name.
        If the account(s) were provided in the input, the account name is always "UNKNOWN".
    """
    if account_id:
        if isinstance(account_id, (str, numbers.Number)):
            account_id_list = [format_account_id(account_id)]
        else:
            account_id_list = [format_account_id(v) for v in account_id]
    else:
        account_id_list = None

    session = botocore.session.Session()

    token_fetcher = get_token_fetcher(session, sso_region, interactive=login, sso_cache=sso_cache)

    token = token_fetcher.fetch_token(start_url)

    config = botocore.config.Config(
        region_name=sso_region,
        signature_version=botocore.UNSIGNED,
    )
    client = session.create_client("sso", config=config)

    if account_id_list:
        def account_iterator():
            for acct in account_id_list:
                yield acct, "UNKNOWN"
    else:
        def account_iterator():
            list_accounts_args = {"accessToken": token["accessToken"]}
            while True:
                response = client.list_accounts(**list_accounts_args)

                for account in response["accountList"]:
                    yield account["accountId"], account["accountName"]

                next_token = response.get("nextToken")
                if not next_token:
                    break
                else:
                    list_accounts_args["nextToken"] = response["nextToken"]

    for account_id, account_name in account_iterator():
        list_role_args = {
            "accessToken": token["accessToken"],
            "accountId": account_id,
        }

        while True:
            response = client.list_account_roles(**list_role_args)

            for role in response["roleList"]:
                role_name = role["roleName"]

                yield account_id, account_name, role_name

            next_token = response.get("nextToken")
            if not next_token:
                break
            else:
                list_role_args["nextToken"] = response["nextToken"]



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import uuid
import random
import string
from collections import namedtuple

"""
instance arn: arn:aws:sso:::instance/ssoins-{16 hex}
identity store: d-{10 hex}
start url: https://d-X.awsapps.com/start
group: {identity-store-id}-uuid
user: {identity-store-id}-uuid
permission set: arn:aws:sso:::permissionSet/{instance-id}/ps-{16 hex}
root: r- [0-9a-z]{4,32}
ou: ou- [0-9a-z]{4,32} - [a-z0-9]{8,32}
account: 123456789012
"""

hex = "0123456789abcdef"
alphanum = string.ascii_lowercase + string.digits

def sample(lst, length):
    return "".join(random.choice(lst) for _ in range(length))

FakeIdentifiers = namedtuple("FakeIdentifiers", [
    "instance_arn",
    "identity_store_id",
    "start_url",
    "principal_id",
    "permission_set_arn",
    "root_id",
    "ou_id",
    "account_id"
])

def generate_fake_identifiers(short_org=False):
    instance_id = "ssoins-{}".format(sample(hex, 16))
    identity_store_id_num = sample(hex, 10)
    identity_store_id = "d-{}".format(identity_store_id_num)

    instance_arn = "arn:aws:sso:::instance/{}".format(instance_id)
    start_url = "https://{}.awsapps.com/start".format(identity_store_id)
    principal_id = "{}-{}".format(identity_store_id_num, uuid.uuid4())
    permission_set_arn = "arn:aws:sso:::permissionSet/{}/ps-{}".format(instance_id, sample(hex, 16))

    root_length = 4 if short_org else random.randint(4, 32)
    root_key = sample(alphanum, root_length)
    ou_length = 8 if short_org else random.randint(8, 32)

    root_id = "r-{}".format(root_key)
    ou_id = "ou-{}-{}".format(root_key, sample(alphanum, ou_length))

    account_id = sample(string.digits, 12)

    return FakeIdentifiers(
        instance_arn,
        identity_store_id,
        start_url,
        principal_id,
        permission_set_arn,
        root_id,
        ou_id,
        account_id
    )

if __name__ == "__main__":
    fake_ids = generate_fake_identifiers()
    data = fake_ids._asdict()
    max_len = max(len(k) for k in data.keys())
    for key, value in data.items():
        print("{} {}".format((key+":").ljust(max_len+1), value))



import numbers

from .lookup import Ids

class FormatError(Exception):
    pass

def format_account_id(account_id):
    if isinstance(account_id, numbers.Number):
        account_id = str(int(account_id))
    if isinstance(account_id, str) and len(account_id) < 12:
        account_id = account_id.rjust(12, "0")
    return account_id

def format_permission_set_arn(ids: Ids, permission_set_id, raise_on_unknown=False):
    if isinstance(permission_set_id, str):
        if permission_set_id.startswith('arn'):
            return permission_set_id
        if permission_set_id.startswith('ssoins-') or permission_set_id.startswith('ins-'):
            return f"arn:aws:sso:::permissionSet/{permission_set_id}"
        if permission_set_id.startswith('ps-'):
            return f"arn:aws:sso:::permissionSet/{ids.instance_id}/{permission_set_id}"
        if raise_on_unknown:
            raise FormatError(f"Unrecognized PermissionSet ID format: {permission_set_id}")
    return permission_set_id



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import logging

import boto3
import aws_error_utils

LOGGER = logging.getLogger(__name__)

from . import format as _format

def _init_cache(cache, key, type):
    if key not in cache:
        cache[key] = type()

class LookupError(Exception):
    pass

class Ids:
    CACHE_KEY_PREFIX = "aws-sso-util-ids-"

    def __init__(self, session, instance_arn=None, identity_store_id=None, cache=None):
        if instance_arn and not instance_arn.startswith("arn:"):
            instance_arn = f"arn:aws:sso:::instance/{instance_arn}"

        if isinstance(session, boto3.Session):
            self._session_fetcher = None
            self._session = session
        elif callable(session):
            self._session_fetcher = session
            self._session = None
        else:
            raise TypeError("Session must be a boto3 Session or a callable returning one")

        self._sso_client = None

        self._instance_arn = instance_arn
        self._instance_arn_specified = (instance_arn is not None)
        self._instance_arn_printed = self._instance_arn_specified

        self._identity_store_id = identity_store_id
        self._identity_store_id_specified = (identity_store_id is not None)
        self._identity_store_id_printed = self._identity_store_id_specified

        self.print_on_fetch = None

        # only cache if we're looking up everything
        if self._instance_arn_specified or self._identity_store_id_specified:
            cache = None
        self._cache = cache
        self._cache_load_attempted = False

    @property
    def session(self):
        if not self._session:
            self._session = self._session_fetcher()
        return self._session

    @property
    def sso_client(self):
        if not self._sso_client:
            self._sso_client = self.session.client("sso-admin")
        return self._sso_client

    def instance_arn_matches(self, instance):
        if not self._instance_arn:
            return True
        if instance and not instance.startswith("arn:"):
            instance = f"arn:aws:sso:::instance/{instance}"
        return instance == self._instance_arn

    @property
    def instance_arn(self):
        if self._instance_arn:
            self._print_instance()
            return self._instance_arn

        if not self._cache_load_attempted:
            success = self._load_from_cache()
            self._cache_load_attempted = True
            if success:
                self._print_instance(cached=True)
                return self._instance_arn

        self._do_lookup("SSO instance", "ARN")
        self._print_instance()
        return self._instance_arn

    @property
    def instance_id(self):
        return self.instance_arn.split("/", 1)[1]

    @property
    def identity_store_id(self):
        if self._identity_store_id:
            self._print_identity_store()
            return self._identity_store_id

        if not self._cache_load_attempted:
            success = self._load_from_cache()
            self._cache_load_attempted = True
            if success:
                self._print_instance(cached=True)
                return self._identity_store_id

        self._do_lookup("identity store", "ID")
        self._print_instance()
        return self._identity_store_id

    def _print(self, message):
        LOGGER.info(message)
        if self.print_on_fetch:
            print(message)

    def _print_instance(self, cached=False):
        if not self._instance_arn_printed:
            cached_str = "cached " if cached else ""
            self._print(f"Using {cached_str}SSO instance {self._instance_arn.split('/')[-1]}")
            self._instance_arn_printed = True

    def _print_identity_store(self, cached=False):
        if not self._identity_store_id_printed:
            cached_str = "cached " if cached else ""
            self._print(f"Using {cached_str}identity store {self._identity_store_id}")
            self._identity_store_id_printed = True

    def _do_lookup(self, lookup_for, identifier):
        instances = self.sso_client.list_instances()["Instances"]
        if len(instances) == 0:
            raise LookupError(f"No {lookup_for} found, please specify {lookup_for} {identifier}")
        if self._instance_arn_specified:
            for instance in instances:
                if self._instance_arn == instance["InstanceArn"]:
                    break
            else:
                raise LookupError(f"No {lookup_for} found matching SSO instance {self._instance_arn}")
        elif self._identity_store_id_specified:
            found_instances = []
            for instance in instances:
                if self._identity_store_id == instance["IdentityStoreId"]:
                    found_instances.append(instance)
            if not found_instances:
                raise LookupError(f"No {lookup_for} found matching identity store id {self._identity_store_id}")
            elif len(found_instances) > 1:
                arns = ", ".join(i["InstanceArn"] for i in found_instances)
                raise LookupError(f"{len(found_instances)} SSO instances found matching identity store id {self._identity_store_id}, please specify SSO instance ARN: {arns}")
            instance = found_instances[0]
        elif len(instances) > 1:
            arns = ", ".join(i["InstanceArn"] for i in instances)
            raise LookupError(f"{len(instances)} SSO instances found, please specify SSO instance ARN: {arns}")
        else:
            instance = instances[0]

        self._instance_arn = instance["InstanceArn"]
        self._identity_store_id = instance["IdentityStoreId"]

        self._store_to_cache()

    def _store_to_cache(self):
        if not self._cache:
            return

        if self.session.profile_name:
            cache_key = f"{self.CACHE_KEY_PREFIX}{self.session.profile_name}"
        else:
            identity = self.session.client("sts").get_caller_identity()
            cache_key = f"{self.CACHE_KEY_PREFIX}{identity['Account']}-{self.session.region_name}"

        self._cache[cache_key] = {
            "InstanceArn": self._instance_arn,
            "IdentityStoreId": self._identity_store_id,
        }

    def _load_from_cache(self):
        if not self._cache:
            return False

        if self.session.profile_name:
            cache_key = f"{self.CACHE_KEY_PREFIX}{self.session.profile_name}"
        else:
            identity = self.session.client("sts").get_caller_identity()
            cache_key = f"{self.CACHE_KEY_PREFIX}{identity['Account']}-{self.session.region_name}"

        if cache_key not in self._cache:
            return False

        instance_arn = self._cache[cache_key].get("InstanceArn")
        identity_store_id = self._cache[cache_key].get("IdentityStoreId")
        if not (instance_arn and identity_store_id):
            return False

        self._instance_arn = instance_arn
        self._identity_store_id = identity_store_id
        return True

_CACHE_KEY_PREFIX_GROUP_ID = "group#id#"
_CACHE_KEY_PREFIX_GROUP_NAME = "group#name#"

def lookup_group_by_id(session: boto3.Session, ids: Ids, group_id, *, cache=None):
    if cache is None:
        cache = {}

    cache_key_id = f"{_CACHE_KEY_PREFIX_GROUP_ID}{group_id}"
    if cache_key_id in cache:
        LOGGER.debug(f"Found group {group_id} in cache")
        group = cache[cache_key_id]
        if isinstance(group, LookupError):
            raise group
        return group

    LOGGER.debug(f"Looking up group {group_id}")

    identity_store = session.client('identitystore')
    try:
        group = identity_store.describe_group(IdentityStoreId=ids.identity_store_id, GroupId=group_id)
        group.pop("ResponseMetadata", None)
    except aws_error_utils.catch_aws_error("ResourceNotFoundException") as e:
        err = LookupError(e)
        cache[cache_key_id] = err
        raise err

    group_name = group["DisplayName"]

    cache_key_name = f"{_CACHE_KEY_PREFIX_GROUP_NAME}{group_name}"
    cache[cache_key_id] = group
    cache[cache_key_name] = group

    return group

def lookup_group_by_name(session: boto3.Session, ids: Ids, group_name, *, cache=None):
    if cache is None:
        cache = {}

    cache_key_name = f"{_CACHE_KEY_PREFIX_GROUP_NAME}{group_name}"
    if cache_key_name in cache:
        LOGGER.debug(f"Found group {group_name} in cache")
        group = cache[cache_key_name]
        if isinstance(group, LookupError):
            raise group
        return group

    LOGGER.debug(f"Looking up group {group_name}")

    identity_store = session.client('identitystore')
    filters=[{'AttributePath': 'DisplayName', 'AttributeValue': group_name}]
    response = identity_store.list_groups(IdentityStoreId=ids.identity_store_id, Filters=filters)

    if len(response['Groups']) == 0:
        err = LookupError("No group named {} found".format(group_name))
        cache[cache_key_name] = err
        raise err
    elif len(response['Groups']) > 1:
        err = LookupError("{} groups named {} found".format(len(response['Groups']), group_name))
        cache[cache_key_name] = err
        raise err
    group = response['Groups'][0]
    group_id = group['GroupId']

    cache_key_id = f"{_CACHE_KEY_PREFIX_GROUP_ID}{group_id}"
    cache[cache_key_id] = group
    cache[cache_key_name] = group

    return group

_CACHE_KEY_PREFIX_USER_ID = "user#id#"
_CACHE_KEY_PREFIX_USER_NAME = "user#name#"

def lookup_user_by_id(session: boto3.Session, ids: Ids, user_id, *, cache=None):
    if cache is None:
        cache = {}

    cache_key_id = f"{_CACHE_KEY_PREFIX_USER_ID}{user_id}"
    if cache_key_id in cache:
        LOGGER.debug(f"Found user {user_id} in cache")
        user = cache[cache_key_id]
        if isinstance(user, LookupError):
            raise user
        return user

    LOGGER.debug(f"Looking up user {user_id}")

    identity_store = session.client('identitystore')
    try:
        user = identity_store.describe_user(IdentityStoreId=ids.identity_store_id, UserId=user_id)
        user.pop("ResponseMetadata", None)
    except aws_error_utils.catch_aws_error("ResourceNotFoundException") as e:
        err = LookupError(e)
        cache[cache_key_id] = err
        raise err

    user_name = user["UserName"]

    cache_key_name = f"{_CACHE_KEY_PREFIX_USER_NAME}{user_name}"
    cache[cache_key_id] = user
    cache[cache_key_name] = user

    return user

def lookup_user_by_name(session: boto3.Session, ids: Ids, user_name, *, cache=None):
    if cache is None:
        cache = {}

    cache_key_name = f"{_CACHE_KEY_PREFIX_USER_NAME}{user_name}"

    if cache_key_name in cache:
        LOGGER.debug(f"Found user {user_name} in cache")
        user = cache[cache_key_name]
        if isinstance(user, LookupError):
            raise user
        return user

    LOGGER.debug(f"Looking up user {user_name}")

    identity_store = session.client('identitystore')
    filters=[{'AttributePath': 'UserName', 'AttributeValue': user_name}]
    response = identity_store.list_users(IdentityStoreId=ids.identity_store_id, Filters=filters)

    if len(response['Users']) == 0:
        err = LookupError("No user named {} found".format(user_name))
        cache[cache_key_name] = err
        raise err
    elif len(response['Users']) > 1:
        err = LookupError("{} users named {} found".format(len(response['Users']), user_name))
        cache[cache_key_name] = err
        raise err

    user = response['Users'][0]
    user_id = user["UserId"]

    cache_key_id = f"{_CACHE_KEY_PREFIX_USER_ID}{user_id}"
    cache[cache_key_id] = user
    cache[cache_key_name] = user

    return user

_CACHE_KEY_PREFIX_PERMISSION_SET_ARN = "ps#arn#"
_CACHE_KEY_PREFIX_PERMISSION_SET_NAME = "ps#name#"

def lookup_permission_set_by_id(session: boto3.Session, ids: Ids, permission_set_id, *, cache=None):
    if cache is None:
        cache = {}

    permission_set_arn = _format.format_permission_set_arn(ids, permission_set_id, raise_on_unknown=True)

    cache_key_arn = f"{_CACHE_KEY_PREFIX_PERMISSION_SET_ARN}{permission_set_arn}"

    if cache_key_arn in cache:
        LOGGER.debug(f"Found permission set {permission_set_id} in cache")
        ps = cache[cache_key_arn]
        if isinstance(ps, LookupError):
            raise ps
        return ps

    LOGGER.debug(f"Looking up permission set {permission_set_id}")

    sso = session.client("sso-admin")

    try:
        ps = sso.describe_permission_set(InstanceArn=ids.instance_arn, PermissionSetArn=permission_set_arn)["PermissionSet"]
    except aws_error_utils.catch_aws_error("ResourceNotFoundException") as e:
        err = LookupError(e)
        cache[cache_key_arn] = err
        raise err

    cache_key_name = f"{_CACHE_KEY_PREFIX_PERMISSION_SET_NAME}{ps['Name']}"

    cache[cache_key_arn] = ps
    cache[cache_key_name] = ps

    return ps

def lookup_permission_set_by_name(session: boto3.Session, ids: Ids, permission_set_name, *, cache=None):
    if cache is None:
        cache = {}

    cache_key_name = f"{_CACHE_KEY_PREFIX_PERMISSION_SET_NAME}{permission_set_name}"

    if cache_key_name in cache:
        LOGGER.debug(f"Found permission set {permission_set_name} in cache")
        ps = cache[cache_key_name]
        if isinstance(ps, LookupError):
            raise ps
        return ps

    LOGGER.debug(f"Looking up permission set {permission_set_name}")

    found_permission_set = None
    sso = session.client("sso-admin")
    paginator = sso.get_paginator('list_permission_sets')
    for ind, response in enumerate(paginator.paginate(InstanceArn=ids.instance_arn)):
        LOGGER.debug(f"ListPermissionSets page {ind+1}: {', '.join(response['PermissionSets'])}")
        for permission_set_arn in response['PermissionSets']:
            ps = sso.describe_permission_set(InstanceArn=ids.instance_arn, PermissionSetArn=permission_set_arn)["PermissionSet"]
            LOGGER.debug(f"PermissionSet {permission_set_arn} has name {ps['Name']}")

            ps_cache_key_arn = f"{_CACHE_KEY_PREFIX_PERMISSION_SET_ARN}{permission_set_arn}"
            ps_cache_key_name = f"{_CACHE_KEY_PREFIX_PERMISSION_SET_NAME}{ps['Name']}"

            cache[ps_cache_key_arn] = ps
            cache[ps_cache_key_name] = ps

            if ps["Name"] == permission_set_name:
                found_permission_set = ps
        if found_permission_set:
            break
    else:
        err = LookupError("No permission set named {} found".format(permission_set_name))
        cache[cache_key_name] = err
        raise err

    return found_permission_set

_CACHE_KEY_PREFIX_ACCOUNT_ID = "account#id#"
_CACHE_KEY_PREFIX_ACCOUNT_NAME = "account#name#"

def lookup_account_by_id(session, account_id, *, cache=None):
    if cache is None:
        cache = {}

    account_id = _format.format_account_id(account_id)

    cache_key_id = f"{_CACHE_KEY_PREFIX_ACCOUNT_ID}{account_id}"

    if cache_key_id in cache:
        LOGGER.debug(f"Found account {account_id} in cache")
        account = cache[cache_key_id]
        if isinstance(account, LookupError):
            raise account
        return account

    LOGGER.debug(f"Looking up account {account_id}")

    organizations = session.client("organizations")
    try:
        account = organizations.describe_account(AccountId=account_id)["Account"]
    except aws_error_utils.catch_aws_error("AccountNotFoundException") as e:
        err = LookupError(e)
        cache[cache_key_id] = err
        raise err

    account_name = account["Name"]

    cache_key_name = f"{_CACHE_KEY_PREFIX_ACCOUNT_NAME}{account_name}"

    cache[cache_key_id] = account
    cache[cache_key_name] = account

    return account

def _acct_str(account):
    if "Name" in account:
        return f"{account['Id']}:{account['Name']}"
    else:
         return f"{account['Id']}"

def lookup_account_by_name(session, account_name, *, cache=None):
    if cache is None:
        cache = {}

    cache_key_name = f"{_CACHE_KEY_PREFIX_ACCOUNT_NAME}{account_name}"

    if cache_key_name in cache:
        LOGGER.debug(f"Found account {account_name} in cache")
        account = cache[cache_key_name]
        if isinstance(account, LookupError):
            raise account
        return account

    LOGGER.debug(f"Looking up account {account_name}")

    found_account = None
    organizations = session.client("organizations")
    paginator = organizations.get_paginator('list_accounts')
    for ind, response in enumerate(paginator.paginate()):
        LOGGER.debug(f"ListAccounts page {ind+1}: {', '.join(a['Name'] for a in response['Accounts'] if 'Name' in a)}")
        for account in response["Accounts"]:
            acct_cache_key_id = f"{_CACHE_KEY_PREFIX_ACCOUNT_ID}{account['Id']}"
            cache[acct_cache_key_id] = account

            if 'Name' in account:
                acct_cache_key_name = f"{_CACHE_KEY_PREFIX_ACCOUNT_NAME}{account['Name']}"
                cache[acct_cache_key_name] = account

            if account.get("Name") == account_name:
                found_account = account
        if found_account:
            break
    else:
        err = LookupError("No account named {} found".format(account_name))
        cache[cache_key_name] = err
        raise err

    return found_account

def lookup_accounts_for_ou(session, ou, *, recursive, refresh=False, cache=None, exclude_org_mgmt_acct=False):
    if cache is None:
        cache = {}

    org_mgmt_acct = False
    if exclude_org_mgmt_acct is True:
        client = session.client("organizations")
        response = client.describe_organization()
        org_mgmt_acct = response["Organization"]["MasterAccountId"]
    elif exclude_org_mgmt_acct:
        org_mgmt_acct = _format.format_account_id(exclude_org_mgmt_acct)

    ou_type = "root" if ou.startswith("r-") else "OU"

    ou_children_key = f"{ou}#children"
    ou_accounts_key = f"{ou}#accounts"

    accounts = []

    if refresh or ou_accounts_key not in cache:
        LOGGER.info(f"Retrieving accounts for {ou_type} {ou}")
        client = session.client("organizations")

        paginator = client.get_paginator("list_accounts_for_parent")
        for ind, response in enumerate(paginator.paginate(ParentId=ou)):
            _init_cache(cache, ou_accounts_key, list)
            if not response["Accounts"]:
                LOGGER.debug(f"No accounts directly in {ou}")
                continue
            acct_strs = [_acct_str(a) for a in response["Accounts"]]
            LOGGER.debug(f"ListAccountsPage page {ind+1} for {ou}: {', '.join(acct_strs)}")
            for account in response["Accounts"]:
                cache[ou_accounts_key].append(account)
                if org_mgmt_acct and account["Id"] == org_mgmt_acct:
                    continue
                yield account
    else:
        acct_strs = [_acct_str(a) for a in cache[ou_accounts_key]]
        if acct_strs:
            LOGGER.debug(f"Loaded cached accounts for {ou_type} {ou}: {', '.join(acct_strs)}")
        else:
            LOGGER.debug(f"Loaded cached accounts for {ou_type} {ou}: (empty list)")
        for account in cache[ou_accounts_key]:
            if org_mgmt_acct and account["Id"] == org_mgmt_acct:
                continue
            yield account

    if recursive:
        if recursive is True:
            child_recursive = True
        else:
            child_recursive = recursive - 1

        if refresh or ou_children_key not in cache:
            LOGGER.info(f"Processing child OUs for {ou_type} {ou}")
            client = session.client("organizations")

            paginator = client.get_paginator("list_organizational_units_for_parent")
            for ind, response in enumerate(paginator.paginate(ParentId=ou)):
                _init_cache(cache, ou_children_key, list)
                if not response["OrganizationalUnits"]:
                    LOGGER.debug(f"No child OUs in {ou}")
                    continue
                sub_ous = [data["Id"] for data in response["OrganizationalUnits"]]
                LOGGER.debug(f"ListOrganizationalUnitsForParent page {ind+1} for {ou}: {', '.join(sub_ous)}")
                for sub_ou_id in sub_ous:
                    cache[ou_children_key].append(sub_ou_id)
                    for account in lookup_accounts_for_ou(session, sub_ou_id,
                            recursive=child_recursive,
                            refresh=refresh,
                            cache=cache,
                            exclude_org_mgmt_acct=org_mgmt_acct):
                        yield account
        else:
            sub_ous = cache[ou_children_key]
            if sub_ous:
                LOGGER.debug(f"Loaded cached child OUs for {ou_type} {ou}: {', '.join(sub_ous)}")
            else:
                LOGGER.debug(f"Loaded cached child OUs for {ou_type} {ou}: (empty list)")
            for sub_ou_id in sub_ous:
                for account in lookup_accounts_for_ou(session, sub_ou_id,
                        recursive=child_recursive,
                        refresh=refresh,
                        cache=cache,
                        exclude_org_mgmt_acct=org_mgmt_acct):
                    yield account

    return accounts




import re
import numbers
import collections
import logging
from collections.abc import Iterable
import itertools

import aws_error_utils

from .lookup import Ids, lookup_accounts_for_ou
from .format import format_account_id

LOGGER = logging.getLogger(__name__)

_Context = collections.namedtuple("_Context", [
    "session",
    "ids",
    "principal",
    "principal_filter",
    "permission_set",
    "permission_set_filter",
    "target",
    "target_filter",
    "get_principal_names",
    "get_permission_set_names",
    "get_target_names",
    "ou_recursive",
    "cache",
    "filter_cache"
])

def _filter(filter_cache, key, func, args):
    if not func:
        return True
    if key not in filter_cache:
        filter_cache[key] = func(*args)
    return filter_cache[key]

def _flatten(list_of_lists):
    return list(itertools.chain(*list_of_lists))

def _is_principal_tuple(principal):
    try:
        return all([
            len(principal) == 2,
            isinstance(principal[0], str),
            principal[0] in ["GROUP", "USER"],
            isinstance(principal[1], str),
        ])
    except:
        return False

def _process_principal(principal):
    if not principal:
        return None
    if isinstance(principal, str):
        return [(None, principal)]
    if _is_principal_tuple(principal):
        return [tuple(principal)]
    else:
        return _flatten(_process_principal(p) for p in principal)

def _process_permission_set(ids, permission_set):
    if not permission_set:
        return None
    if not isinstance(permission_set, str) and isinstance(permission_set, Iterable):
        return _flatten(_process_permission_set(ids, ps) for ps in permission_set)

    if permission_set.startswith("arn"):
        permission_set_arn = permission_set
    elif permission_set.startswith("ssoins-") or permission_set.startswith("ins-"):
        permission_set_arn = f"arn:aws:sso:::permissionSet/{permission_set}"
    elif permission_set.startswith("ps-"):
        permission_set_arn = f"arn:aws:sso:::permissionSet/{ids.instance_id}/{permission_set}"
    else:
        raise TypeError(f"Invalid permission set id {permission_set}")
    return [permission_set_arn]

def _is_target_tuple(target):
    try:
        return all([
            len(target) == 2,
            isinstance(target[0], str),
            target[0] in ["AWS_OU", "AWS_ACCOUNT"],
            isinstance(target[1], str),
        ])
    except:
        return False

def _process_target(target):
    if not target:
        return None
    if isinstance(target, numbers.Number):
        return [("AWS_ACCOUNT", format_account_id(target))]
    if isinstance(target, str):
        if re.match(r"^\d+$", target):
            return [("AWS_ACCOUNT", format_account_id(target))]
        elif re.match(r"^r-[a-z0-9]{4,32}$", target) or re.match(r"^ou-[a-z0-9]{4,32}-[a-z0-9]{8,32}$", target):
            return [("AWS_OU", target)]
        else:
            raise TypeError(f"Invalid target {target}")
    elif _is_target_tuple(target):
        target_type, target_id = target
        if target_type not in ["AWS_ACCOUNT", "AWS_OU"]:
            raise TypeError(f"Invalid target type {target_type}")
        return [(target_type, target_id)]
    else:
        value = _flatten(_process_target(t) for t in target)
        return value

def _get_account_iterator(target, context: _Context):
    def target_iterator():
        target_name = None
        if context.get_target_names:
            organizations_client = context.session.client("organizations")
            account = organizations_client.describe_account(AccountId=target[1])["Account"]
            if account.get("Name"):
                target_name = account["Name"]
        value = (*target, target_name)
        if not _filter(context.filter_cache, value[1], context.target_filter, value):
            LOGGER.debug(f"Account is filtered: {value}")
        else:
            LOGGER.debug(f"Visiting single account: {value}")
            yield value
    return target_iterator

def _get_ou_iterator(target, context: _Context):
    def target_iterator():
        target_name = None
        # if context.get_target_names:
        #     organizations_client = context.session.client("organizations")
        #     ou = organizations_client.describe_organizational_unit(OrganizationalUnitId=target[1])["OrganizationalUnit"]
        #     if ou.get("Name"):
        #         target_name = ou("Name")
        value = (*target, target_name)
        accounts = lookup_accounts_for_ou(context.session, value[1], recursive=context.ou_recursive)
        for account in accounts:
            yield "AWS_ACCOUNT", account["Id"], account["Name"]
    return target_iterator

def _get_single_target_iterator(target, context: _Context):
    target_type = target[0]
    if target_type == "AWS_ACCOUNT":
        return _get_account_iterator(target, context)
    elif target_type == "AWS_OU":
        return _get_ou_iterator(target, context)
    else:
        raise TypeError(f"Invalid target type {target_type}")

def _get_all_accounts_iterator(context: _Context):
    def target_iterator():
        organizations_client = context.session.client("organizations")
        accounts_paginator = organizations_client.get_paginator("list_accounts")
        for response in accounts_paginator.paginate():
            LOGGER.debug(f"ListAccounts page: {response}")
            for account in response["Accounts"]:
                account_id = account["Id"]
                account_name = account["Name"]

                value = ("AWS_ACCOUNT", account_id, account_name)

                if not _filter(context.filter_cache, account_id, context.target_filter, value):
                    LOGGER.debug(f"Account is filtered: {value}")
                    continue

                LOGGER.debug(f"Visiting account: {value}")
                yield value
    return target_iterator

def _get_target_iterator(context: _Context):
    if context.target:
        iterables = [_get_single_target_iterator(t, context) for t in context.target]
        def target_iterator():
            return itertools.chain(*[it() for it in iterables])
        return target_iterator
    else:
        LOGGER.debug(f"Iterating for all accounts")
        return _get_all_accounts_iterator(context)

def _get_single_permission_set_iterator(permission_set, context: _Context):
    permission_set_arn = permission_set
    permission_set_id = permission_set_arn.split("/")[-1]

    def permission_set_iterator(target_type, target_id, target_name):
        if not context.get_permission_set_names:
            permission_set_name = None
        else:
            sso_admin_client = context.session.client("sso-admin")
            response = sso_admin_client.describe_permission_set(
                InstanceArn=context.ids.instance_arn,
                PermissionSetArn=permission_set_arn
            )
            LOGGER.debug(f"DescribePermissionSet response: {response}")
            permission_set_name = response["PermissionSet"]["Name"]

        if not _filter(context.filter_cache, permission_set_arn, context.permission_set_filter, (permission_set_arn, permission_set_name)):
            LOGGER.debug(f"Single permission set is filtered: {(permission_set_id, permission_set_name)}")
        else:
            LOGGER.debug(f"Visiting single permission set {(permission_set_id, permission_set_name)}")
            yield permission_set_arn, permission_set_id, permission_set_name
    return permission_set_iterator

def _get_all_permission_sets_iterator(context: _Context):
    def permission_set_iterator(target_type, target_id, target_name):
        if target_type != "AWS_ACCOUNT":
            raise TypeError(f"Unsupported target type {target_type}")
        sso_admin_client = context.session.client("sso-admin")
        permission_sets_paginator = sso_admin_client.get_paginator("list_permission_sets_provisioned_to_account")
        for response in permission_sets_paginator.paginate(
                InstanceArn=context.ids.instance_arn,
                AccountId=target_id):
            LOGGER.debug(f"ListPermissionSetsProvisionedToAccount {target_id} page: {response}")
            if "PermissionSets" not in response:
                continue

            for permission_set_arn in response["PermissionSets"]:
                permission_set_id = permission_set_arn.split("/", 2)[-1]
                if not context.get_permission_set_names:
                    permission_set_name = None
                else:
                    if permission_set_arn not in context.cache:
                        response = sso_admin_client.describe_permission_set(
                            InstanceArn=context.ids.instance_arn,
                            PermissionSetArn=permission_set_arn
                        )
                        LOGGER.debug(f"DescribePermissionSet response: {response}")
                        context.cache[permission_set_arn] = response["PermissionSet"]["Name"]
                    permission_set_name = context.cache[permission_set_arn]

                if not _filter(context.filter_cache, permission_set_arn, context.permission_set_filter, (permission_set_arn, permission_set_name)):
                    LOGGER.debug(f"Permission set is filtered: {(permission_set_id, permission_set_name)}")
                    continue

                LOGGER.debug(f"Visiting permission set: {(permission_set_id, permission_set_name)}")
                yield permission_set_arn, permission_set_id, permission_set_name
    return permission_set_iterator

def _get_permission_set_iterator(context: _Context):
    if context.permission_set:
        iterables = [_get_single_permission_set_iterator(ps, context) for ps in context.permission_set]
        def permission_set_iterator(target_type, target_id, target_name):
            return itertools.chain(*[it(target_type, target_id, target_name) for it in iterables])
        return permission_set_iterator
    else:
        LOGGER.debug("Iterating for all permission sets")
        return _get_all_permission_sets_iterator(context)

def _get_principal_iterator(context: _Context):
    def principal_iterator(
            target_type, target_id, target_name,
            permission_set_arn, permission_set_id, permission_set_name):
        if target_type != "AWS_ACCOUNT":
            raise TypeError(f"Unsupported target type {target_type}")

        sso_admin_client = context.session.client("sso-admin")
        identity_store_client = context.session.client("identitystore")

        assignments_paginator = sso_admin_client.get_paginator("list_account_assignments")
        for response in assignments_paginator.paginate(
                InstanceArn=context.ids.instance_arn,
                AccountId=target_id,
                PermissionSetArn=permission_set_arn):
            LOGGER.debug(f"ListAccountAssignments for {target_id} {permission_set_arn.split('/')[-1]} page: {response}")

            if not response["AccountAssignments"] and not "NextToken" in response:
                LOGGER.debug(f"No assignments for {target_id} {permission_set_arn.split('/')[-1]}")

            for assignment in response["AccountAssignments"]:
                principal_type = assignment["PrincipalType"]
                principal_id = assignment["PrincipalId"]
                LOGGER.debug(f"Visiting principal {principal_type}:{principal_id}")

                if context.principal:
                    for principal in context.principal:
                        type_matches = (principal[0] is None or principal[0] != principal_type)
                        if type_matches and principal[1] == principal_id:
                            LOGGER.debug(f"Found principal {principal_type}:{principal_id}")
                            break
                    else:
                        LOGGER.debug(f"Principal {principal_type}:{principal_id} does not match principals")
                        continue

                principal_key = (principal_type, principal_id)
                if not context.get_principal_names:
                    principal_name = None
                else:
                    if principal_key not in context.cache:
                        if principal_type == "GROUP":
                            try:
                                response = identity_store_client.describe_group(
                                    IdentityStoreId=context.ids.identity_store_id,
                                    GroupId=principal_id
                                )
                                LOGGER.debug(f"DescribeGroup response: {response}")
                                context.cache[principal_key] = response["DisplayName"]
                            except aws_error_utils.catch_aws_error("ResourceNotFoundException"):
                                context.cache[principal_key] = None
                        elif principal_type == "USER":
                            try:
                                response = identity_store_client.describe_user(
                                    IdentityStoreId=context.ids.identity_store_id,
                                    UserId=principal_id
                                )
                                LOGGER.debug(f"DescribeUser response: {response}")
                                context.cache[principal_key] = response["UserName"]
                            except aws_error_utils.catch_aws_error("ResourceNotFoundException"):
                                context.cache[principal_key] = None
                        else:
                            raise ValueError(f"Unknown principal type {principal_type}")
                    principal_name = context.cache[principal_key]

                if not _filter(context.filter_cache, principal_key, context.principal_filter, (principal_type, principal_id, principal_name)):
                    if context.principal:
                        LOGGER.debug(f"Principal is filtered: {principal_type}:{principal_id}")
                    else:
                        LOGGER.debug(f"Principal is filtered: {principal_type}:{principal_id}")
                    continue

                LOGGER.debug(f"Visiting principal: {principal_type}:{principal_id}")
                yield principal_type, principal_id, principal_name
    return principal_iterator

Assignment = collections.namedtuple("Assignment", [
    "instance_arn",
    "principal_type",
    "principal_id",
    "principal_name",
    "permission_set_arn",
    "permission_set_name",
    "target_type",
    "target_id",
    "target_name",
])

def list_assignments(
        session,
        instance_arn=None,
        identity_store_id=None,
        principal=None,
        principal_filter=None,
        permission_set=None,
        permission_set_filter=None,
        target=None,
        target_filter=None,
        get_principal_names=False,
        get_permission_set_names=False,
        get_target_names=False,
        ou_recursive=False):
    """Iterate over AWS SSO assignments.

    Args:
        session (boto3.Session): boto3 session to use
        instance_arn (str): The SSO instance to use, or it will be looked up using ListInstances
        identity_store_id (str): The identity store to use if principal names are being retrieved
            or it will be looked up using ListInstances
        principal: A principal specification or list of principal specifications.
            A principal specification is a principal id or a 2-tuple of principal type and id.
        principal_filter: A callable taking principal type, principal id, and principal name
            (which may be None), and returning True if the principal should be included.
        permission_set: A permission set arn or id, or a list of the same.
        permission_set_filter: A callable taking permission set arn and name (name may be None),
            returning True if the permission set should be included.
        target: A target specification or list of target specifications.
            A target specification is an account or OU id, or a 2-tuple of target type, which
            is either AWS_ACCOUNT or AWS_OU, and target id.
        target_filter: A callable taking target type, target id, and target name
            (which may be None), and returning True if the target should be included.
        get_principal_names (bool): Retrieve names for principals in assignments.
        get_permission_set_names (bool): Retrieve names for permission sets in assignments.
        get_target_names (bool): Retrieve names for targets in assignments.
        ou_recursive (bool): Set to True if an OU is provided as a target to get all accounts
            including those in child OUs.

    Returns:
        An iterator over Assignment namedtuples
    """
    ids = Ids(lambda: session, instance_arn, identity_store_id)

    return _list_assignments(
        session,
        ids,
        principal=principal,
        principal_filter=principal_filter,
        permission_set=permission_set,
        permission_set_filter=permission_set_filter,
        target=target,
        target_filter=target_filter,
        get_principal_names=get_principal_names,
        get_permission_set_names=get_permission_set_names,
        get_target_names=get_target_names,
        ou_recursive=ou_recursive,
    )

def _list_assignments(
        session,
        ids,
        principal=None,
        principal_filter=None,
        permission_set=None,
        permission_set_filter=None,
        target=None,
        target_filter=None,
        get_principal_names=False,
        get_permission_set_names=False,
        get_target_names=False,
        ou_recursive=False):

    principal = _process_principal(principal)
    permission_set = _process_permission_set(ids, permission_set)
    target = _process_target(target)

    cache = {}
    filter_cache = {}

    context = _Context(
        session = session,
        ids=ids,
        principal=principal,
        principal_filter=principal_filter,
        permission_set=permission_set,
        permission_set_filter=permission_set_filter,
        target=target,
        target_filter=target_filter,
        get_principal_names=get_principal_names,
        get_permission_set_names=get_permission_set_names,
        get_target_names=get_target_names,
        ou_recursive=ou_recursive,
        cache=cache,
        filter_cache=filter_cache,
    )

    target_iterator = _get_target_iterator(context)

    permission_set_iterator = _get_permission_set_iterator(context)

    principal_iterator = _get_principal_iterator(context)

    for target_type, target_id, target_name in target_iterator():
        for permission_set_arn, permission_set_id, permission_set_name, in permission_set_iterator(target_type, target_id, target_name):
            for principal_type, principal_id, principal_name in principal_iterator(
                    target_type, target_id, target_name,
                    permission_set_arn, permission_set_id, permission_set_name):

                assignment = Assignment(
                    ids.instance_arn,
                    principal_type,
                    principal_id,
                    principal_name,
                    permission_set_arn,
                    permission_set_name,
                    target_type,
                    target_id,
                    target_name,
                )
                LOGGER.debug(f"Visiting assignment: {assignment}")
                yield assignment

if __name__ == "__main__":
    import boto3
    import sys
    import json

    logging.basicConfig(level=logging.INFO)

    kwargs = {}
    for v in sys.argv[1:]:
        if hasattr(logging, v):
            LOGGER.setLevel(getattr(logging, v))
        else:
            kwargs = json.loads(v)

    def fil(*args):
        print(args)
        return True

    kwargs["target_filter"] = fil

    try:
        session = boto3.Session()
        print(",".join(Assignment._fields))
        for value in list_assignments(session, **kwargs):
            print(",".join(v or "" for v in value))
    except KeyboardInterrupt:
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aws_sso_lib', 'aws_sso_lib.vendored_botocore']

package_data = \
{'': ['*']}

install_requires = \
['aws-error-utils>=1.0.4,<2.0.0', 'boto3>=1.14.0']

setup_kwargs = {
    'name': 'aws-sso-lib',
    'version': '1.5.0',
    'description': 'Library to make AWS SSO easier',
    'long_description': '# `aws-sso-lib`\n\n`aws-sso-lib` allows you to programmatically interact with AWS SSO.\n\nThe primary functions that will be of interest are available at the package level:\n* `get_boto3_session`: Get a boto3 session for a specific account and role.\n* `login`: ensure the user is logged in to AWS SSO, with dispatch to the browser.\n* `list_available_accounts` and `list_available_roles`: discover the access the user has.\n* `list_assignments`: for admin purposes, iterate over all assignments in AWS SSO, which is currently hard to do through the API.\n\n`aws-sso-util` is a command-line utility built on `aws-sso-lib` for interacting with AWS SSO; see the details of that project [here](https://github.com/benkehoe/aws-sso-util).\n\n## Install\n\n```\npip install --user aws-sso-lib\npython -c "import aws_sso_lib; aws_sso_lib.login(\'https://my-start-url.awsapps.com/start\', \'us-east-2\')"\n```\n\n## `get_boto3_session`\n\nOften when writing a script, you know the exact account and role you want the script to use.\nYou could configure a profile in your `~/.aws/config` for this (perhaps using `aws-sso-util configure profile`), but especially if multiple people may be using the script, it\'s more convenient to have the configuration baked into the script itself.\n`get_boto3_session()` is the function to do that with.\n\n```python\nget_boto3_session(start_url, sso_region, account_id, role_name, region, login=False)\n```\n\n* `start_url`: [REQUIRED] The start URL for the AWS SSO instance.\n* `sso_region`: [REQUIRED] The AWS region for the AWS SSO instance.\n* `account_id`: [REQUIRED] The AWS account ID to use.\n* `role_name`: [REQUIRED] The AWS SSO role (aka PermissionSet) name to use.\n* `region`: [REQUIRED] The AWS region for the boto3 session.\n* `login`: Set to `True` to interactively log in the user if their AWS SSO credentials have expired.\n* Returns a [boto3 Session object](https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html) configured for the account and role.\n\nFor more control over the login process, use the `login` function separately.\n\n## `login`\n\nWhile the functions that require the user to be logged in let you specify `login=True` to interactively log in the user if they are not already logged in, you can have more control over the process, or retrieve the access token, using `login()`.\n\nIf the user is not logged in or `force_refresh` is `True`, it will attempt to log in.\nIf the user is logged in and `force_refresh` is `False`, no action is taken.\n\nNormally, it will attempt to automatically open the user\'s browser to log in, as well as printing the URL and code to stderr as a fallback. However, if `disable_browser` is `True`, or if `disable_browser` is `None` (the default) and the environment variable `AWS_SSO_DISABLE_BROWSER` is set to `1` or `true`, only the message with the URL and code will be printed.\n\nA custom message can be printed by setting `message` to a template string using `{url}` and `{code}` as placeholders.\nThe message can be suppressed by setting `message` to `False`.\n\n```python\nlogin(start_url, sso_region, force_refresh=False, disable_browser=None, message=None, outfile=None)\n```\n\n* `start_url`: [REQUIRED] The start URL for the AWS SSO instance.\n* `sso_region`: [REQUIRED] The AWS region for the AWS SSO instance.\n* `force_refresh`: Set to `True` to always go through the authentication process.\n* `disable_browser`: Set to `True` to skip the browser popup and only print a message with the URL and code.\n* `message`: A message template to print with the fallback URL and code, or `False` to suppress the message.\n* `outfile`: The file-like object to print the message to (stderr by default)\n* Returns the token dict as returned by [sso-oidc:CreateToken](https://docs.aws.amazon.com/singlesignon/latest/OIDCAPIReference/API_CreateToken.html), which contains the actual authorization token, as well as the expiration.\n\n## `list_available_accounts` and `list_available_roles`\n\nAWS SSO provides programmatic access to the permissions that a user has.\nYou can access this through `list_available_accounts()` and `list_available_roles()`.\n\nWith both, you can set `login=True` to interactively log in the user if they are not already logged in.\n\nNote that these functions return iterators; they don\'t return a list, because the number of roles may be very large and you shouldn\'t have to wait for the entire list to be created to start processing.\nYou can always get a list by, for example, `list(list_available_roles(...))`.\n\n```python\nlist_available_accounts(start_url, sso_region, login=False)\n```\n\n* `start_url`: The start URL for the AWS SSO instance.\n* `sso_region`: The AWS region for the AWS SSO instance.\n* `login`: Set to `True` to interactively log in the user if their AWS SSO credentials have expired.\n* Returns an iterator that yields account id and account name.\n\n```\nlist_available_roles(start_url, sso_region, account_id=None, login=False)\n```\n\n* `start_url`: [REQUIRED] The start URL for the AWS SSO instance.\n* `sso_region`: [REQUIRED] The AWS region for the AWS SSO instance.\n* `account_id`: Optional account id or list of account ids to check.\n  * If not set, all accounts available to the user are used.\n* `login`: Set to `True` to interactively log in the user if their AWS SSO credentials have expired.\n* Returns an iterator that yields account id, account name, and role name.\n\n## `list_assignments`\n\nThe AWS SSO API only allows you to [list assignments for a specific account _and_ permission set](https://docs.aws.amazon.com/singlesignon/latest/APIReference/API_ListAccountAssignments.html).\nTo find all your assignments, you need to iterate over all accounts, and then interate over all permission sets.\n`list_assignments()` does this work for you.\n\nUnlike the other functions list above, this uses admin APIs, which require AWS credentials, rather than taking as input a start URL and region.\n\n`list_assignments` returns an iterator over `Assignment` named tuples, which have the following fields:\n\n* `instance_arn`\n* `principal_type`\n* `principal_id`\n* `principal_name`\n* `permission_set_arn`\n* `permission_set_name`\n* `target_type`\n* `target_id`\n* `target_name`\n\nThe name fields may be `None`, if the names are not known or looked up.\nBy default, principal and permission set names are not retrieved, nor are account names for accounts that have been provided as explicit targets.\n\nIf you don\'t specify `instance_arn` and/or `identity_store_id`, these will be looked up using the [ListInstances API](https://docs.aws.amazon.com/singlesignon/latest/APIReference/API_ListInstances.html), which today returns at most one instance (with associated identity store).\n\nAn assignment is the combination of a principal (a user or a group), a permission set, and a target (an AWS account).\nFor each of these values, you can provide either an explicit specification, or a filter function.\n\nYou can provide an OU as a target, which will use all accounts in that OU, and optionally all accounts recursively in child OUs as well.\n\n```python\nlist_assignments(\n    session,\n    instance_arn=None,\n    identity_store_id=None,\n    principal=None,\n    principal_filter=None,\n    permission_set=None,\n    permission_set_filter=None,\n    target=None,\n    target_filter=None,\n    get_principal_names=False,\n    get_permission_set_names=False,\n    get_target_names=False,\n    ou_recursive=False)\n```\n\n* `session`: [REQUIRED] boto3 session to use\n* `instance_arn`: The SSO instance to use, or it will be looked up using ListInstances\n* `identity_store_id`: The identity store to use if principal names are being retrieved or it will be looked up using ListInstances\n* `principal`: A principal specification or list of principal specifications.\n  * A principal specification is a principal id or a 2-tuple of principal type and id.\n* `principal_filter`: A callable taking principal type, principal id, and principal name (which may be `None`), and returning `True` if the principal should be included.\n* `permission_set`: A permission set arn or id, or a list of the same.\n* `permission_set_filter`: A callable taking permission set arn and name (name may be `None`), returning True if the permission set should be included.\n* `target`: A target specification or list of target specifications.\n  * A target specification is an account or OU id, or a 2-tuple of target type, which is either AWS_ACCOUNT or AWS_OU, and target id.\n* `target_filter`: A callable taking target type, target id, and target name (which may be `None`), and returning `True` if the target should be included.\n* `get_principal_names`: Set to `True` to retrieve names for principals in assignments.\n* ` get_permission_set_names`: Set to `True` to retrieve names for permission sets in assignments.\n* `get_target_names`: Set to `True` to retrieve names for targets in assignments, when they are explicitly provided as targets. For OUs as targets or if no targets are specified, the account names will be retrieved automatically during the enumeration process.\n* `ou_recursive`: Set to `True` if an OU is provided as a target to get all accounts including those in child OUs.\n* Returns an iterator over `Assignment` tuples\n',
    'author': 'Ben Kehoe',
    'author_email': 'ben@kehoe.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/benkehoe/aws-sso-util',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
